<?php

declare (strict_types=1);
namespace ScopedGtbabel\Dotenv;

use ScopedGtbabel\Dotenv\Exception\InvalidPathException;
use ScopedGtbabel\Dotenv\Loader\Loader;
use ScopedGtbabel\Dotenv\Loader\LoaderInterface;
use ScopedGtbabel\Dotenv\Parser\Parser;
use ScopedGtbabel\Dotenv\Parser\ParserInterface;
use ScopedGtbabel\Dotenv\Repository\Adapter\ArrayAdapter;
use ScopedGtbabel\Dotenv\Repository\Adapter\PutenvAdapter;
use ScopedGtbabel\Dotenv\Repository\RepositoryBuilder;
use ScopedGtbabel\Dotenv\Repository\RepositoryInterface;
use ScopedGtbabel\Dotenv\Store\StoreBuilder;
use ScopedGtbabel\Dotenv\Store\StoreInterface;
use ScopedGtbabel\Dotenv\Store\StringStore;
class Dotenv
{
    /**
     * The store instance.
     *
     * @var \Dotenv\Store\StoreInterface
     */
    private $store;
    /**
     * The parser instance.
     *
     * @var \Dotenv\Parser\ParserInterface
     */
    private $parser;
    /**
     * The loader instance.
     *
     * @var \Dotenv\Loader\LoaderInterface
     */
    private $loader;
    /**
     * The repository instance.
     *
     * @var \Dotenv\Repository\RepositoryInterface
     */
    private $repository;
    /**
     * Create a new dotenv instance.
     *
     * @param \Dotenv\Store\StoreInterface           $store
     * @param \Dotenv\Parser\ParserInterface         $parser
     * @param \Dotenv\Loader\LoaderInterface         $loader
     * @param \Dotenv\Repository\RepositoryInterface $repository
     *
     * @return void
     */
    public function __construct(\ScopedGtbabel\Dotenv\Store\StoreInterface $store, \ScopedGtbabel\Dotenv\Parser\ParserInterface $parser, \ScopedGtbabel\Dotenv\Loader\LoaderInterface $loader, \ScopedGtbabel\Dotenv\Repository\RepositoryInterface $repository)
    {
        $this->store = $store;
        $this->parser = $parser;
        $this->loader = $loader;
        $this->repository = $repository;
    }
    /**
     * Create a new dotenv instance.
     *
     * @param \Dotenv\Repository\RepositoryInterface $repository
     * @param string|string[]                        $paths
     * @param string|string[]|null                   $names
     * @param bool                                   $shortCircuit
     * @param string|null                            $fileEncoding
     *
     * @return \Dotenv\Dotenv
     */
    public static function create(\ScopedGtbabel\Dotenv\Repository\RepositoryInterface $repository, $paths, $names = null, bool $shortCircuit = \true, string $fileEncoding = null)
    {
        $builder = $names === null ? \ScopedGtbabel\Dotenv\Store\StoreBuilder::createWithDefaultName() : \ScopedGtbabel\Dotenv\Store\StoreBuilder::createWithNoNames();
        foreach ((array) $paths as $path) {
            $builder = $builder->addPath($path);
        }
        foreach ((array) $names as $name) {
            $builder = $builder->addName($name);
        }
        if ($shortCircuit) {
            $builder = $builder->shortCircuit();
        }
        return new self($builder->fileEncoding($fileEncoding)->make(), new \ScopedGtbabel\Dotenv\Parser\Parser(), new \ScopedGtbabel\Dotenv\Loader\Loader(), $repository);
    }
    /**
     * Create a new mutable dotenv instance with default repository.
     *
     * @param string|string[]      $paths
     * @param string|string[]|null $names
     * @param bool                 $shortCircuit
     * @param string|null          $fileEncoding
     *
     * @return \Dotenv\Dotenv
     */
    public static function createMutable($paths, $names = null, bool $shortCircuit = \true, string $fileEncoding = null)
    {
        $repository = \ScopedGtbabel\Dotenv\Repository\RepositoryBuilder::createWithDefaultAdapters()->make();
        return self::create($repository, $paths, $names, $shortCircuit, $fileEncoding);
    }
    /**
     * Create a new mutable dotenv instance with default repository with the putenv adapter.
     *
     * @param string|string[]      $paths
     * @param string|string[]|null $names
     * @param bool                 $shortCircuit
     * @param string|null          $fileEncoding
     *
     * @return \Dotenv\Dotenv
     */
    public static function createUnsafeMutable($paths, $names = null, bool $shortCircuit = \true, string $fileEncoding = null)
    {
        $repository = \ScopedGtbabel\Dotenv\Repository\RepositoryBuilder::createWithDefaultAdapters()->addAdapter(\ScopedGtbabel\Dotenv\Repository\Adapter\PutenvAdapter::class)->make();
        return self::create($repository, $paths, $names, $shortCircuit, $fileEncoding);
    }
    /**
     * Create a new immutable dotenv instance with default repository.
     *
     * @param string|string[]      $paths
     * @param string|string[]|null $names
     * @param bool                 $shortCircuit
     * @param string|null          $fileEncoding
     *
     * @return \Dotenv\Dotenv
     */
    public static function createImmutable($paths, $names = null, bool $shortCircuit = \true, string $fileEncoding = null)
    {
        $repository = \ScopedGtbabel\Dotenv\Repository\RepositoryBuilder::createWithDefaultAdapters()->immutable()->make();
        return self::create($repository, $paths, $names, $shortCircuit, $fileEncoding);
    }
    /**
     * Create a new immutable dotenv instance with default repository with the putenv adapter.
     *
     * @param string|string[]      $paths
     * @param string|string[]|null $names
     * @param bool                 $shortCircuit
     * @param string|null          $fileEncoding
     *
     * @return \Dotenv\Dotenv
     */
    public static function createUnsafeImmutable($paths, $names = null, bool $shortCircuit = \true, string $fileEncoding = null)
    {
        $repository = \ScopedGtbabel\Dotenv\Repository\RepositoryBuilder::createWithDefaultAdapters()->addAdapter(\ScopedGtbabel\Dotenv\Repository\Adapter\PutenvAdapter::class)->immutable()->make();
        return self::create($repository, $paths, $names, $shortCircuit, $fileEncoding);
    }
    /**
     * Create a new dotenv instance with an array backed repository.
     *
     * @param string|string[]      $paths
     * @param string|string[]|null $names
     * @param bool                 $shortCircuit
     * @param string|null          $fileEncoding
     *
     * @return \Dotenv\Dotenv
     */
    public static function createArrayBacked($paths, $names = null, bool $shortCircuit = \true, string $fileEncoding = null)
    {
        $repository = \ScopedGtbabel\Dotenv\Repository\RepositoryBuilder::createWithNoAdapters()->addAdapter(\ScopedGtbabel\Dotenv\Repository\Adapter\ArrayAdapter::class)->make();
        return self::create($repository, $paths, $names, $shortCircuit, $fileEncoding);
    }
    /**
     * Parse the given content and resolve nested variables.
     *
     * This method behaves just like load(), only without mutating your actual
     * environment. We do this by using an array backed repository.
     *
     * @param string $content
     *
     * @throws \Dotenv\Exception\InvalidFileException
     *
     * @return array<string,string|null>
     */
    public static function parse(string $content)
    {
        $repository = \ScopedGtbabel\Dotenv\Repository\RepositoryBuilder::createWithNoAdapters()->addAdapter(\ScopedGtbabel\Dotenv\Repository\Adapter\ArrayAdapter::class)->make();
        $phpdotenv = new self(new \ScopedGtbabel\Dotenv\Store\StringStore($content), new \ScopedGtbabel\Dotenv\Parser\Parser(), new \ScopedGtbabel\Dotenv\Loader\Loader(), $repository);
        return $phpdotenv->load();
    }
    /**
     * Read and load environment file(s).
     *
     * @throws \Dotenv\Exception\InvalidPathException|\Dotenv\Exception\InvalidEncodingException|\Dotenv\Exception\InvalidFileException
     *
     * @return array<string,string|null>
     */
    public function load()
    {
        $entries = $this->parser->parse($this->store->read());
        return $this->loader->load($this->repository, $entries);
    }
    /**
     * Read and load environment file(s), silently failing if no files can be read.
     *
     * @throws \Dotenv\Exception\InvalidEncodingException|\Dotenv\Exception\InvalidFileException
     *
     * @return array<string,string|null>
     */
    public function safeLoad()
    {
        try {
            return $this->load();
        } catch (\ScopedGtbabel\Dotenv\Exception\InvalidPathException $e) {
            // suppressing exception
            return [];
        }
    }
    /**
     * Required ensures that the specified variables exist, and returns a new validator object.
     *
     * @param string|string[] $variables
     *
     * @return \Dotenv\Validator
     */
    public function required($variables)
    {
        return (new \ScopedGtbabel\Dotenv\Validator($this->repository, (array) $variables))->required();
    }
    /**
     * Returns a new validator object that won't check if the specified variables exist.
     *
     * @param string|string[] $variables
     *
     * @return \Dotenv\Validator
     */
    public function ifPresent($variables)
    {
        return new \ScopedGtbabel\Dotenv\Validator($this->repository, (array) $variables);
    }
}
